import VirtualElement from './VirtualElement'
import DefaultDOMElement from '../dom/DefaultDOMElement'

const COMPONENT_ROOT_ELEMENT = Symbol('COMPONENT_ROOT_ELEMENT')

// TODO: rewrite RenderingEngine with all new requirements in mind
// and with an implementation backed by a documented algorithm description.
export default class RenderingEngine {
  constructor (options = {}) {
    this.componentFactory = options.componentFactory
    if (!this.componentFactory) throw new Error("'componentFactory' is mandatory")
    this.elementFactory = options.elementFactory || DefaultDOMElement.createDocument('html')
    if (!this.elementFactory) throw new Error("'elementFactory' is mandatory")
  }

  // TODO: rethink this
  // it would make more sense to work with newProps, newState
  // than with oldProps, oldState as in the current implementation
  updateComponent (comp, newProps, newState, state) {
    // let oldProps = comp.props
    // let oldState = comp.state
    state = state || this._createState()
    // NOTE: don't ask shouldRerender if no element is there yet
    let needRerender = !comp.el || comp.shouldRerender(newProps, newState)
    // in any case update props, triggering willReceiveProps
    comp._setProps(newProps)

    if (needRerender) {
      let el = this._update(comp.el, comp.render.bind(comp), {
        mode: COMPONENT_ROOT_ELEMENT,
        owner: comp,
        context: comp.context
      })
      if (!comp.el) {
        comp.el = el
      }
    }
  }

  _update (el, render, { mode, owner, context, htmlProps, state }) {
    let renderingContext = new CaptureContext(owner)

    let virtualEl = render(renderingContext.$$)

    // Validity Check: the returned virtual element must either be
    // - a virtual HTML element, created such as `$$('div')`
    // - or a virtual Component
    if (!virtualEl) {
      throw new Error('Component.render() returned nil.')
    }
    if (!virtualEl._isVirtualHTMLElement && !virtualEl._isVirtualComponent) {
      throw new Error('render() must return a plain element or a Component')
    }

    // Forwarding Components: leaving a flag that the virtual element should be treated as a forwarded component
    if (mode === COMPONENT_ROOT_ELEMENT && virtualEl._isVirtualComponent) {
      virtualEl._isForwarded = true
      virtualEl.parent = owner
    }

    // merge HTML attributes and properties set by owner into the virtualElement
    // e.g. if owner does something like this: `$$(Foo).addClass('se-foo'), then this class needs to be retained when rerendering Foo itself.
    if (htmlProps) {
      virtualEl._mergeHTMLConfig(htmlProps)
    }

    // TODO: explain why this is necessary
    if (mode === COMPONENT_ROOT_ELEMENT && !virtualEl._isVirtualComponent) {
      virtualEl._comp = owner
    }
  }
}

class CaptureContext {
  constructor (owner) {
    this.owner = owner
    this.refs = {}
    this.foreignRefs = {}
    this.elements = []
    this.components = []
    this.$$ = this._$$.bind(this)
    this.$$.capturing = true
  }

  _$$ () {
    let vel = VirtualElement.createElement.apply(this, arguments)
    vel._context = this
    vel._owner = this.owner
    if (vel._isVirtualComponent) {
      // virtual components need to be captured recursively
      this.components.push(vel)
    }
    this.elements.push(vel)
    return vel
  }
}
